--[[
	$Id: vActionbar.lua 17 2013-01-14 15:29:51Z frafall $
	
	vActionbar - manage actionbars based on 2nd class as well as 1st class
                 and enable custom action bars by command, If no save/load name
				 given it will default to PrimarySecondary class.
				 
				 /abar save [<name>]
				 /abar load [<name>]
				 /abar list [<name>]

	By Frafall@Isiltir
	
	Thx to ehysta for his post on skill saving/restoring which is the basis for this addon.
	Ref http://forum.us.runesofmagic.com/showthread.php?t=69888&p=520482&viewfull=1#post520482
	
	TODO:
	- Equipped items 
	- Set skills 
	- Pet skills priest pet
	Arcadia
--]]
vActionbar = {}
_G.vActionbar = vActionbar;

vActionbarDebug = false

local function msg(txt, r, g, b)
	DEFAULT_CHAT_FRAME:AddMessage(tostring(txt), (r or 1), (g or 1), (b or 0))
end

local function vprint(txt, ...)
	if vActionbarDebug then
		msg(string.format(txt, unpack(arg)), 1, 1, 0)
	end
end

function pack(...)
	return arg
end

function map(t)
	local keys={}
	local n=0
	for k,v in pairs(t) do
		n=n+1
		keys[n]=k
	end
	table.sort(keys, function(a,b) return a>b end)
	return keys
end

--
-- Timed functions
--
local timedFunction = {}

-- Make this a stack, not a hash?
-- Or static, we only ever have 1 outstanding fx (for now?)
local function RegisterTimedFunction(fx, delay, args)
	args = args or {}
	timedFunction[fx] = {active = true, delay = delay, args = args}
end

--
-- Manage sound when doing the cursor stuff
--

local function RestoreAudio(args)
    InterfaceSFXVolumeSlider_SetValue(args.sfx or 100)
end
 
local function MuteAudio()
    local sfx = InterfaceSFXVolumeSlider_GetValue()
    InterfaceSFXVolumeSlider_SetValue(0)
end

--
-- Implementation
--
local T_EMPTY     = 0
local T_SKILL     = 1
local T_MACRO     = 2
local T_ITEM      = 3
local T_WORN      = 4
local T_PARTNER   = 5 
local T_TRANSPORT = 6
local T_SETSKILL  = 7

-- Get tooltip text
local function getTooltip(side, index)
	local lineobj = _G["GameTooltipText"..side..index]
	if lineobj then
		local text = lineobj:GetText()
		lineobj:SetText("")
		return text
	end
	return ""
end

--
-- List the actions
--
local function ListActions(actions)
	for slot = 1,80 do
		if actions[slot] and actions[slot].name then
			msg(string.format("%02d - %s", slot, actions[slot].name))
		end
	end
end

--
-- Save the actions
--
local function GetActions()

	--  skills table
    local skills = {}
	
	-- build skills table
    local function lookupSkills(page)
        local si = 1
        repeat
            local name,i1,icon,passive,rank,i2,cost,skillable,avail = GetSkillDetail(page,si)			
			if icon then icon = icon:gsub("\\","/"):lower() end
            if not name then break end
            local link = GetSkillHyperLink(page,si)
            local sid = link:match("^|H.-:(%x+)") or ''		
            local skill = { ['name'] = name, ['icon'] = icon, ['id'] = tonumber(sid,10), ['page'] = page, ['index'] = si }
            skills[name] = skill
            si = si + 1
        until name == nil
    end

	function getID(slot)
		local id = 0
		
		PickupAction(slot)
		if CursorItemType() == 'action' then
			id = GetCursorItemInfo() - 1
		end
		PickupAction(slot)
		
		return id
	end

	lookupSkills(1) -- generic
	lookupSkills(2) -- mainclass
	lookupSkills(3) -- subclass
    lookupSkills(4) -- elite
	lookupSkills(6) -- Titel Skills
	local verb = vActionbarDebug
	local actions = {}
	for slot=1,80 do 
		local icon, name, count, locked, wore, cont = GetActionInfo(slot)
		if icon then
			icon = icon:gsub("\\","/"):lower()
			vprint("%02d: %s, %s, %s, %s, %s, %s",slot, tostring(icon), tostring(name), tostring(count), tostring(locked), tostring(wore), tostring(cont))
		end
		GameTooltip.SetActionItem(GameTooltip,  slot)
		
		-- Empty link
		if name == nil then
			local skill = { ['type'] = T_EMPTY }
			actions[slot] = skill 
			
		-- Link to set-skill
		elseif name == '' and wore ~= nil and cont ~= nil and icon:match("interface/icons/skill_panel_icons/.*") then
			local id = getID(slot)			
			local name = string.match(getTooltip("Left", 1), "([^\+]+).*$")
			name = name:gsub(" $", "")
			local skill = { ['type'] = T_SETSKILL, name = name, id = id }			
			vprint("%02d: set-skill %s [%d]", slot, skill.name, skill.id)
			actions[slot] = skill
			
		-- Link to skill pages 1-4
		elseif name == '' and wore ~= nil and cont ~= nil then
			local id = getID(slot) 
			local skill = nil
			local name = string.match(getTooltip("Left", 1), "([^\+]+).*$")
			name = name:gsub(" $", "")
			
			if skills[name] then
				skill = skills[name]
			else
				msg(string.format("%02d: skill does not exist for this class \"%s\"", slot, name), 1, 0, 0)
			end
			
			if skill then
				vprint("%02d: skill %s [%d/%d]", slot, skill.name, skill.page, skill.index)
				skill.type = T_SKILL
				actions[slot] = skill 
			end

		-- Link to transport point
		elseif name ~= '' and wore == nil and cont == nil and icon:match("interface/icons/map_zone.*") then
			local id = getID(slot) + 1
			local skill = { ['type'] = T_TRANSPORT, name = name, slot = id }
			vprint("%02d: transport point %s [%d]", slot, skill.name, skill.slot)
			actions[slot] = skill
						
		-- Link to partner mount (did they mis-spell Spirit Shartooth on purpose (moumt_siya)
		elseif name ~= '' and wore == nil and cont == nil and icon:match("mou.t_.*") then
			local skill = { ['type'] = T_PARTNER, name = name, page = 2 }
			vprint("%02d: mount %s [%d]", slot, skill.name, skill.page)
			actions[slot] = skill
			
		-- Link to partner pet
		elseif name ~= '' and wore == nil and cont == nil and icon:match("pet_") then
			local skill = { ['type'] = T_PARTNER, name = name, page = 1 }
			vprint("%02d: pet %s [%d]", slot, skill.name, skill.page)
			actions[slot] = skill
		
		-- Link to macro, note the ID is some internal order, havent figured it out
		elseif name ~= '' and wore == nil and cont == nil then
			local id = getID(slot)
			local skill = { ['type'] = T_MACRO, name = name, slot = id }
			vprint("%02d: macro %s [%d]", slot, skill.name, skill.slot)
			actions[slot] = skill
		
		-- Link to item worn 
		elseif name == '' and wore == true and cont == nil then
			local name = string.match(getTooltip("Left", 1), "([^\+]+).*$")
			name = name:gsub(" $", "")
			local id = getID(slot) 
			local skill = { ['type'] = T_WORN, itemid = id, name = name }
			vprint("%02d: worn \"%s\"", slot, skill.name)
			actions[slot] = skill

		-- Link to item in backpack
		elseif name == '' and wore == false and cont == nil then
			local id = getID(slot) 
			local skill = { ['type'] = T_ITEM, itemid = id }
			vprint("%02d: item [%d]", slot, skill.itemid)
			actions[slot] = skill

		-- Unknown actionbar link
		else
			msg(string.format("%02d: Unknown actionbar link", slot), 1, 0, 0)
		end
	end
	
	GameTooltip:Hide();
	return actions
end

local function SetActions(actions)
	local baghash = {}
	local wornhash = {}
	local partnerhash = {}
	local suithash = {}
	local macrohash = {}
	
	
	function setActionSlot(delay, keys)
		local slot = table.remove(keys)
		local action = actions[slot]
		local verb = true				
	
		-- Skill page/index given
		if action.type == T_SKILL then
			vprint("%02d: skill %s %d:%d",slot, action.name, action.page, action.index)
			DragSkillButton(action.page, action.index)
			PickupAction(slot)
			DeleteCursorItem()

		-- Set skill
		elseif action.type == T_SETSKILL then
			vprint("%02d: set-skill \"%s\" %d",slot, action.name, action.id)
			if(suithash[action.name]) then
				SkillPlateDragStart(suithash[action.name])	
				PickupAction(slot)
				DeleteCursorItem()
			else
				msg(string.format("%02d: unknown set skill id \"%s\"", slot, action.name), 1, 0, 0)
			end
		
		-- Macro slot number given, need to verify existence macrohash[macro_name]
		elseif action.type == T_MACRO then
			vprint("%02d: macro %s",slot, action.name)
			if macrohash[action.name] then
				PickupMacroItem(macrohash[action.name])
				PickupAction(slot)
				DeleteCursorItem()
			else
				msg(string.format("%02d: unknown macro \"%s\"", slot, action.name), 1, 0, 0)
			end

		-- Transport point
		elseif action.type == T_TRANSPORT then
			vprint("%02d: transport %s %d",slot, action.name, action.slot)
			TB_PickupItem(action.slot)
			PickupAction(slot)
			DeleteCursorItem()

			
		-- Item ID in bag given, need to search bag to find
		elseif action.type == T_ITEM then
			if baghash[action.itemid] then
				vprint("%02d: item: %s [%d]",slot, baghash[action.itemid].name, baghash[action.itemid].slot)
				PickupBagItem(GetBagItemInfo(baghash[action.itemid].slot))
				PickupAction(slot)
				DeleteCursorItem()
			--[[
			else
				msg(string.format("%02d: unknown item id %d", slot, action.itemid), 1, 0, 0)
			--]]
			end
			
		-- Item ID in worn equipment given, need to search 
		elseif action.type == T_WORN then			
			vprint("%02d: worn item \"%s\"",slot, action.name)
			if wornhash[action.name] then 
				PickupEquipmentItem(wornhash[action.name].slot)
				PickupAction(slot)
				DeleteCursorItem()
			else
				msg(string.format("%02d: unknown worn item \"%s\"", slot, action.name), 1, 0, 0)
			end

			
		-- Partners
		elseif action.type == T_PARTNER then
			vprint("%02d: partner %d \"%s\"", slot, action.page, action.name)
			PartnerFrame_PickPartner(action.page, partnerhash[action.page][action.name])
			PickupAction(slot)
			DeleteCursorItem()

		-- Empty
		elseif action.type == T_EMPTY then
			--vprint("%02d: -empty-", slot)
			PickupAction(slot)
			DeleteCursorItem()
			
		else
			vprint("%02d: unknown",slot)
		end
		
		-- If finished, Unmute Audio else Prepare to be called again
		if table.getn(keys) > 0 then
			RegisterTimedFunction(setActionSlot, delay, pack(delay, keys))
		else
			--msg("Done!")
			local args = { ['sfx'] = sfx }
			RegisterTimedFunction(RestoreAudio, 1, args)
		end
	end
	
	-- Preload macro names/slots - doesnt handle duplicate names..
	for i = 1, 56 do
		local icon_index, macro_name, macro_body = GetMacroInfo(i)
		if macro_name then
			macrohash[macro_name] = i
		end
	end
	
	-- Preload the suit skill name/index
	for i = 0, SkillPlateUpdate(-1) - 1, 1 do
		local name, texture = SkillPlateUpdate(i);	
		if ( name ) then
			suithash[name] = i
		end
	end
	
	-- Preload item id and slot number from gear worn -----HIER EDITIEREN 17 before 23 ist artefact
	for wornIndex = 1,23 do
		GameTooltip:SetInventoryItem("player", wornIndex)
		local name = string.match(getTooltip("Left", 1), "([^\+]+).*$")
		name = name:gsub(" $", "")
		wornhash[name] = { slot = wornIndex, name = name }
	end
	GameTooltip:Hide();
	
	-- Preload the partner name/slot to partnerhash
	for mode=1,2 do
		partnerhash[mode] = {}
		for idx=1,PartnerFrame_GetPartnerCount(mode) do
			local icon,dbid,name,c1,c2,c3,c4=PartnerFrame_GetPartnerInfo(mode,idx)
			partnerhash[mode][name] = idx
		end
	end
	
	local BagSlotCount,BagSize,MaxBagSlots = GetBagCount()
	-- Preload item id and slot number from backpack
	for bagIndex = 1,BagSize do
		local inventoryIndex, icon, name, itemCount, locked, invalid = GetBagItemInfo(bagIndex)
		local link = GetBagItemLink(inventoryIndex)
		if link then
			local sid = (link and tonumber(link:match("^|H.-:(%x+)"),16)) or ''
			baghash[sid] = { slot = bagIndex, name = name }	
		end
	end
	
	-- We dont want audio while doing this
	MuteAudio()
	
	-- Wait 1s before starting to avoid crashing w. built-in stuff
	RegisterTimedFunction(setActionSlot, 1, pack(0.005, map(actions)))
end

--
-- Key for class change saves
--
local function getcid()
	local mc,sc = UnitClassToken("player")
	return string.gsub(string.lower(mc),"^%l", string.upper) .. string.gsub(string.lower(sc),"^%l", string.upper)
end

--
-- Public API
--
function vSaveActions(name)
	name = name or getcid()
	MuteAudio()
	vActionbar[name] = GetActions()	
end

local loading = false

function vLoadActions(name)
	name = name or getcid()
	if vActionbar[name] then
		SetActions(vActionbar[name])
	else
		msg(string.format("Action set '%s' does not exist", name), 1, 0, 0)
	end
end

function vListActionSets(name)
	if name and vActionbar[name] then
		msg("Actionbar for "..name..":")
		ListActions(vActionbar[name])
	else
		msg("Available actionbar sets:")
		for name, actions in pairs(vActionbar) do
			msg("   "..name)
		end
	end
end

--
-- RoM Addon API
--

-- Slash handler
local function vActionbar_Slash(cmd, param)
	function splitWords(str)
		local t = {}
		local function helper(word) table.insert(t, word) return "" end
		if not str:gsub("%w+", helper):find"%S" then return t end
	end
	
	local cmd = nil
	local file 
	local sargs
	
	if param then
		sarg = splitWords(param)
		
		if #sarg >= 1 then
			cmd = sarg[1]:lower()
		end
		if #sarg >= 2 then
			file = sarg[2]
		end
	end
	
	if cmd == "save" then
		file = file or getcid()
		msg("Saving the actionbar to "..file)
		vSaveActions(file)
		
	elseif cmd == "load" then
		file = file or getcid()
		msg("Loading the actionbar from "..file)
		vLoadActions(file)
		
	elseif cmd == "list" then
		vListActionSets(file)
		
	elseif cmd == "debug" then
		vActionbarDebug = not vActionbarDebug
		msg(string.format("Debugging is %s", tostring(vActionbarDebug)))
		
	else
		msg("Use /abar [save/load/list] <filename>")
		msg("If no filename is given, classes is used")
	end
end

-- Load handler
function vActionbar_OnLoad(this)
	this:RegisterEvent("EXCHANGECLASS_SHOW")
	this:RegisterEvent("EXCHANGECLASS_SUCCESS")

	SLASH_vActionbar_Slash1 = "/abar"
	SlashCmdList["vActionbar_Slash"] = vActionbar_Slash
	
	SaveVariablesPerCharacter("vActionbar")
end

-- Delayed execution functions
function vActionbar_OnUpdate(elapsed)
	for fx, v in pairs(timedFunction) do
		if v.active then
			v.delay = v.delay - elapsed
			if v.delay <= 0 then
				timedFunction[fx].active = false
				pcall(fx, unpack(v.args))	
			end
		end
	end
end

function vActionbar_OnEvent(frame, event)
	if event == "EXCHANGECLASS_SHOW" then
		if not IsShiftKeyDown() then
			SendSystemMsg("SAVED ACTIONBAR")
			vSaveActions()
		end
		
	elseif event == "EXCHANGECLASS_SUCCESS" then	
		    if not IsShiftKeyDown() then              --if arg1 == "player" then
			print("LOADED ACTIONBAR")
			vLoadActions()
		end
	end
end
